--[[
# Element: Offline Indicator

Handles the visibility and updating of an Indicator based on the unit is offline.

## Widget

OfflineIndicator - A `Texture` used to display if the unit is offline.

## Notes

A default texture will be applied if the widget is a Texture and doesn't have a texture or a color set.

## Examples

    -- Position and size
    local OfflineIndicator = self:CreateTexture(nil, 'OVERLAY')
    OfflineIndicator:SetSize(16, 16)
    OfflineIndicator:SetPoint('TOPRIGHT', self)

    -- Register it with oUF
    self.OfflineIndicator = OfflineIndicator
--]]

local _, ns = ...
local oUF = ns.oUF

local Update = function(self, event, unit)
	if(self.unit ~= unit) then return end
	local element = self.OfflineIndicator
	
	local isOffline = not UnitIsConnected(unit)
	if(isOffline) then
		element:Show()
	else
		element:Hide()
	end

	if(element.PostUpdate) then
		return element:PostUpdate(isOffline)
	end
end

local Path = function(self, ...)
	return (self.OfflineIndicator.Override or Update)(self, ...)
end

local ForceUpdate = function(element)
	return Path(element.__owner, 'ForceUpdate')
end

local Enable = function(self, unit)
	local element = self.OfflineIndicator
	if(element) then
		element.__owner = self
		element.ForceUpdate = ForceUpdate

		self:RegisterEvent('UNIT_FLAGS', Path)
		self:RegisterEvent("UNIT_CONNECTION", Path)

		if(element:IsObjectType('Texture') and not element:GetTexture()) then
			element:SetTexture([[Interface\CharacterFrame\Disconnect-Icon]])
			element:SetTexCoord(.08, .92, .08, .92)
		end

		return true
	end
end

local Disable = function(self)
	local element = self.OfflineIndicator

	if(element) then
		self:UnregisterEvent('UNIT_FLAGS', Path)
		self:UnregisterEvent("UNIT_CONNECTION", Path)
	end
end

oUF:AddElement("OfflineIndicator", Path, Enable, Disable)