local _, ns = ...
local oUF = ns.oUF

-- Lua APIs
local pairs = pairs

-- WoW APIs
local CreateFrame = CreateFrame
local UnitGUID = UnitGUID
local UnitAura = UnitAura
local UnitIsUnit = UnitIsUnit
local GetSpellInfo = GetSpellInfo

-- Register Libs
local LCD = LibStub('LibClassicDurations', true)

local GUIDs = {}
local PLAYER_UNITS = {
    player = true,
    pet = true,
}

local SetupGUID
do
    local cache = setmetatable({}, {__type = 'k'})
    
    local frame = CreateFrame('Frame')
    frame:SetScript('OnEvent', function(self, event)
        for k, t in pairs(GUIDs) do
            GUIDs[k] = nil
            for a in pairs(t) do
                t[a] = nil
            end
            cache[t] = true
        end
    end)
    frame:RegisterEvent('PLAYER_REGEN_ENABLED')
    frame:RegisterEvent('PLAYER_ENTERING_WORLD')
    
    function SetupGUID(guid)
        local t = next(cache)
        if t then
            cache[t] = nil
        else
            t = {}
        end
        GUIDs[guid] = t
    end
end

local function ResetIcon(watch, icon, count, duration, expiration)
    if (icon.onlyShowMissing) then
        icon:Hide()
    else
        icon:Show()
        if (icon.cd) then
            if duration and (duration > 0) then
                icon.cd:SetCooldown(expiration - duration, duration)
                icon.cd:Show()
            else
                icon.cd:Hide()
            end
        end
        if (icon.count) then
            icon.count:SetText(count > 1 and count)
        end
        if (icon.overlay) then
            icon.overlay:Hide()
        end
        icon:SetAlpha(icon.presentAlpha)
        icon.icon:SetDesaturated(false)
    end
end

local function ExpireIcon(watch, icon)
    if (icon.onlyShowPresent) then
        icon:Hide()
    else
        icon:Show()
        if (icon.cd) then icon.cd:Hide() end
        if (icon.count) then icon.count:SetText() end
        if (icon.overlay) then icon.overlay:Show() end
        icon:SetAlpha(icon.missingAlpha)
        icon.icon:SetDesaturated(true)
    end
end

local found = {}
local function Update(self, event, unit)
    if (self.unit ~= unit) then return end
    local watch = self.AuraWatch
    local index, icons = 1, watch.watched
    local _, name, texture, count, duration, expiration, caster, key, icon, spellID
    local filter = 'HELPFUL'
    local guid = UnitGUID(unit)
    if not guid then return end
    if not GUIDs[guid] then SetupGUID(guid) end
    
    for key, icon in pairs(icons) do
        icon:Hide()
    end
    
    while true do
        if LCD and not UnitIsUnit('player', unit) then
            name, texture, count, _, duration, expiration, caster, _, _, spellID = LCD:UnitAura(unit, index, filter)
        else
            name, texture, count, _, duration, expiration, caster, _, _, spellID = UnitAura(unit, index, filter)
        end
        
        if (not name) then
            if (filter == 'HELPFUL') then
                filter = 'HARMFUL'
                index = 1
            else
                break
            end
        else
            if (watch.strictMatching) then
				key = spellID
			else
				key = name..texture
			end
			icon = icons[key]
            if icon and (icon.anyUnit or (caster and icon.fromUnits and icon.fromUnits[caster])) then
                ResetIcon(watch, icon, count, duration, expiration)
                GUIDs[guid][key] = true
                found[key] = true
            end
            index = index + 1
        end
    end
    
    for key in pairs(GUIDs[guid]) do
        if icons[key] and not found[key] then
            ExpireIcon(watch, icons[key])
        end
    end
    
    for k in pairs(found) do
		found[k] = nil
	end
end

local function SetupIcons(self)
    local watch = self.AuraWatch
    local icons = watch.icons
    watch.watched = {}
    
    for _, icon in pairs(icons) do
        local name, _, texture = GetSpellInfo(icon.spellID)
        if (name) then
            icon.name = name
            
            if (not icon.cd) and not (watch.hideCooldown or icon.hideCooldown) then
				local cd = CreateFrame('Cooldown', nil, icon, 'CooldownFrameTemplate')
                cd:SetAllPoints(icon)
                cd.noCooldownCount = true
                icon.cd = cd
			end

			if (not icon.icon) then
				local tex = icon:CreateTexture(nil, 'BACKGROUND')
				tex:SetAllPoints(icon)
				tex:SetTexture(texture)
                tex:SetTexCoord(.08, .92, .08, .92)
				icon.icon = tex
                if (not icon.overlay) then
					local overlay = icon:CreateTexture(nil, 'OVERLAY')
					overlay:SetTexture([[Interface\Buttons\UI-Debuff-Overlays]])
					overlay:SetAllPoints(icon)
					overlay:SetTexCoord(.296875, .5703125, 0, .515625)
					overlay:SetVertexColor(1, 0, 0)
					icon.overlay = overlay
				end
			end

			if (not icon.count) and not (watch.hideCount or icon.hideCount) then
				local count = icon:CreateFontString(nil, 'OVERLAY')
				count:SetFontObject(NumberFontNormal)
				count:SetPoint('BOTTOMRIGHT', icon, 'BOTTOMRIGHT', -1, 0)
				icon.count = count
			end
            
            if (icon.onlyShowMissing == nil) then
                icon.onlyShowMissing = watch.onlyShowMissing
            end
            if (icon.onlyShowPresent == nil) then
                icon.onlyShowPresent = watch.onlyShowPresent
            end
            if (icon.fromUnits == nil) then
                icon.fromUnits = watch.fromUnits or PLAYER_UNITS
            end
            if (icon.anyUnit == nil) then
                icon.anyUnit = watch.anyUnit
            end
            if (watch.strictMatching) then
				watch.watched[icon.spellID] = icon
			else
				watch.watched[name..texture] = icon
			end
            
            watch.watched[icon.spellID] = icon
        else
            print('oUF_AuraWatch error: spell ID ' .. tostring(icon.spellID) .. ' no longer exists!')
        end
    end
end

local function Enable(self)
	if self.AuraWatch then
		self:RegisterEvent('UNIT_AURA', Update)
		SetupIcons(self)
        
		return true
	else
		return false
	end
end

local function Disable(self)
	if self.AuraWatch then
		self:UnregisterEvent('UNIT_AURA', Update)
		for _, icon in pairs(self.AuraWatch.icons) do
			icon:Hide()
		end
	end
end

oUF:AddElement('AuraWatch', Update, Enable, Disable)
