﻿local _, ns = ...
local oUF = ns.oUF
local cfg = ns.cfg
local filter = ns.filter
local castbar = ns.castbar
local aurabar = ns.aurabar

-- Lua APIs
local unpack 	= unpack
local pairs 	= pairs
local ceil 	 	= math.ceil
local floor		= math.floor
local huge	 	= math.huge
local format	= string.format
local tsort 	= table.sort
local tinsert 	= table.insert

-- WoW APIs
local CreateFrame 		= CreateFrame
local GetTime 			= GetTime
local UnitAura 			= UnitAura
local UnitClass   		= UnitClass
local UnitIsUnit		= UnitIsUnit
local UnitIsDead 		= UnitIsDead
local UnitIsGhost 		= UnitIsGhost
local UnitIsFriend 		= UnitIsFriend
local UnitPowerType		= UnitPowerType
local UnitIsConnected 	= UnitIsConnected
local UnitIsFeignDeath  = UnitIsFeignDeath
local UnitIsDeadOrGhost = UnitIsDeadOrGhost

-- Register Libs
local LCG = LibStub('LibCustomGlow-1.0')
local LCD = LibStub('LibClassicDurations')

-- player class
local playerClass = select(2, UnitClass('player'))

-- global short numbers
function ShortValue(num)
    if (num >= 1e6) then
		return format('%.1fm', num / 1e6)
    elseif (num >= 1e3) then
        return format('%.1fk', num / 1e3)
    else
        return num
    end
end

-- global format time
function FormatTime(sec)
	local Day, Hour, Minute = 86400, 3600, 60

	if (sec == huge) then
		sec = 0
	end

	if (sec >= Day) then
		return format('%dd', ceil(sec / Day))
	elseif (sec >= Hour) then
		return format('%dh', ceil(sec / Hour))
	elseif (sec >= Minute) then
		return format('%dm', ceil(sec / Minute))
	elseif (sec >= Minute / 12) then
		return floor(sec)
	end

	return format('%.1f', sec)
end

-- change power bar colors
local Powercolor = setmetatable({
    power = setmetatable({
        ['MANA'] 	= {	0.12, 0.56, 1.0	 },
        ['RAGE'] 	= {	0.91, 0.26, 0.26 },
		['ENERGY'] 	= {	0.85, 0.83, 0.15 },
    }, {__index = oUF.colors.power}),
}, {__index = oUF.colors})

-- health update
local PostUpdateHealth = function(Health, unit, cur, max)
	local self = Health.__owner
	local percent = floor(cur / max * 100 + .5)

	-- status strings
	if (not UnitIsConnected(unit)) then
		self.curhp:Hide()
		self.Health:SetValue(0)
		self.perhp:SetText('|cffFF0000' .. 'OFF' .. '|r')
	elseif (UnitIsFeignDeath(unit)) then
		self.curhp:Hide()
		self.Health:SetValue(max)
		self.perhp:SetText('|cffFFCC00' .. 'F.Dead' .. '|r')
	elseif (UnitIsDead(unit) and not UnitIsFeignDeath(unit)) then
		self.curhp:Hide()
		self.Health:SetValue(0)
		self.perhp:SetText('|cffFF3700' .. 'Dead' .. '|r')
	elseif (UnitIsGhost(unit)) then
		self.curhp:Hide()
		self.Health:SetValue(0)
		self.perhp:SetText('|cffFF3333' .. 'GHO' .. '|r')
	elseif ((unit == 'player' or unit == 'target' or unit:match('party%d?$')) and (cur ~= max)) then
		self.curhp:Show()
		self.perhp:SetText(percent)
	else
		self.curhp:Hide()
		self.perhp:SetText(percent)
	end
	
    -- set arrow percent colors
	if (percent < 20) then
		self.Arrow:SetVertexColor(1, 0.2, 0.2)
        self.perhp:SetTextColor(1, 0.2, 0.2)
	elseif (percent < 40) then
		self.Arrow:SetVertexColor(1, 0.5, 0)
        self.perhp:SetTextColor(1, 0.5, 0)
	elseif (percent < 70) then
		self.Arrow:SetVertexColor(1, 0.9, 0)
        self.perhp:SetTextColor(1, 0.9, 0)
	else
		self.Arrow:SetVertexColor(1, 1, 1)
        self.perhp:SetTextColor(1, 1, 1)
	end

	-- curhp position
	if (percent > 50) then
		self.curhp:SetPoint('TOPRIGHT', self.Health:GetStatusBarTexture(), 'BOTTOMRIGHT', 5, 5)
	else
		self.curhp:SetPoint('TOPRIGHT', self.Health:GetStatusBarTexture(), 'BOTTOMRIGHT', 20, 5)
	end

	-- hide hparrow
	if ((unit ~= 'player') or (cur == max) or UnitIsDeadOrGhost(unit)) then
		self.Arrow:Hide()
	else
		self.Arrow:Show()
	end
end

-- power update
local PostUpdatePower = function(Power, unit, cur, min, max)
	local self = Power.__owner
	local percent = floor(cur / max * 100 + .5)
	
	-- status curpp
	if (not UnitIsConnected(unit)) then
		self.curpp:Hide()
		self.Power:SetValue(0)
	elseif (UnitIsDeadOrGhost(unit)) then
		self.curpp:Hide()
		self.Power:SetValue(0)
	elseif ((unit == 'player' or unit == 'target') and (cur ~= 0) and (cur ~= max)) then
		self.curpp:Show()
	else
		self.curpp:Hide()
	end
	
	-- curpp position
	if (percent > 50) then
		self.curpp:SetPoint('TOPRIGHT', self.Power:GetStatusBarTexture(), 'BOTTOMRIGHT', 5, 2)
	else
		self.curpp:SetPoint('TOPRIGHT', self.Power:GetStatusBarTexture(), 'BOTTOMRIGHT', 20, 2)
	end
end

-- power update inraid
local PostUpdatePowerRaid = function(Power, unit, cur, min, max)
	local self = Power.__owner
	local _, powerType = UnitPowerType(unit)
	
	-- status powerbar inraid
	if (UnitIsDeadOrGhost(unit)) then
		self.Power:Hide()
	elseif (not UnitIsConnected(unit)) then
		self.Power:Hide()
	elseif ((powerType == 'MANA' or powerType == 'RAGE') and (cur ~= 0) and (cur ~= max)) then
		self.Power:Show()
	else
		self.Power:Hide()
	end

	-- hide curpp inraid
	self.curpp:Hide()
end	

-- dispellist
local CanDispel = {
	PRIEST  = { Magic  = true, 	Disease = true  },
    SHAMAN  = { Poison = true, 	Disease = true  },
    PALADIN = { Magic  = true,	Poison  = true,	Disease = true	},
    DRUID   = { Curse  = true,	Poison  = true	},
    MAGE    = { Curse  = true	},
	WARLOCK = { Magic  = true	},
}
local DispelList = CanDispel[playerClass] or {}

-- mouseover zoom auras for hook
local function OnEnterAura(self, button)
    self.HighlightAura:Show()
    self.HighlightAura:SetPoint('CENTER', button)
    self.HighlightAura:SetSize(button:GetWidth() * cfg.auraZoom, button:GetHeight() * cfg.auraZoom)
    self.HighlightAura:SetFrameLevel(button:GetFrameLevel() + 1)
    
    self.HighlightAura.button:SetTexCoord(.08, .92, .08, .92)
    self.HighlightAura.button:SetTexture(button.icon:GetTexture())
    self.HighlightAura.button:SetSize(button:GetWidth() * cfg.auraZoom, button:GetHeight() * cfg.auraZoom)
    
    self.HighlightAura.overlay:SetSize(button:GetWidth() * cfg.auraZoom + 15, button:GetHeight() * cfg.auraZoom + 15)
    self.HighlightAura.overlay:SetVertexColor(button.overlay:GetVertexColor())
end

local function OnLeaveAura(self)
    self.HighlightAura:Hide()
end

-- auras timer
local CreateAuraTimer = function(self, elapsed)
	if (self.expiration) then
		self.expiration = self.expiration - elapsed
		-- set auras fade anim
		if (self.isPlayer and self.expiration > 0 and self.expiration < 5) then
			if (not self.Fader:IsPlaying()) then
				self.Fader:Play()
			end
		else
			if (self.Fader:IsPlaying()) then
				self.Fader:Stop()
			end
		end
		-- set text color
		if (self.expiration > 0 and self.expiration < cfg.auraTimer) then
			self.time:SetFormattedText(FormatTime(self.expiration))

			if (self.expiration < 5) then
				self.time:SetTextColor(1, .3, .3)
			elseif (self.expiration < 10) then
				self.time:SetTextColor(1, 1, 0)
			else
				self.time:SetTextColor(unpack(cfg.timecolor))
			end
		else
			self:SetScript('OnUpdate', nil)
			self.time:SetText()
		end
	end
end

-- create auras icon
local PostCreateIcon = function(auras, button)
	auras.disableCooldown 		= cfg.hideCDs
	button.cd.noOCC 			= cfg.hideOcc
    button.cd.noCooldownCount 	= cfg.hideCDc
	
	button.icon:SetTexCoord(.08, .92, .08, .92)
	
    button.overlay:SetTexture(cfg.Auratex)
    button.overlay:SetPoint('TOPLEFT', button, -6, 6)
    button.overlay:SetPoint('BOTTOMRIGHT', button, 6, -6)
    button.overlay:SetTexCoord(0, 1, 0, 1)
	button.overlay.Hide = function(self) self:SetVertexColor(unpack(cfg.maincolor)) end
	
	local StringParent = CreateFrame('Frame', nil, button)
	StringParent:SetFrameLevel(button:GetFrameLevel() + 1)
	
    button.time = StringParent:CreateFontString(nil, 'OVERLAY')
	button.time:SetFont(cfg.AuraFont, cfg.LargeFS, cfg.FontF)
	button.time:SetPoint('TOPLEFT', button, 1, 4)
	button.time:SetJustifyH('LEFT')
    button.time:SetTextColor(unpack(cfg.timecolor))
    
	button.count:SetParent(StringParent)
	button.count:SetFont(cfg.AuraFont, cfg.LargeFS, cfg.FontF)
    button.count:SetPoint('BOTTOMRIGHT', button, 2, -1)
	button.count:SetJustifyH('RIGHT')
    button.count:SetTextColor(unpack(cfg.countcolor))
	
	button.Fader = button:CreateAnimationGroup()
	button.Fader:SetLooping('BOUNCE')

	button.Fader.Anim = button.Fader:CreateAnimation('Alpha')
	button.Fader.Anim:SetFromAlpha(1)
	button.Fader.Anim:SetToAlpha(0.5)
	button.Fader.Anim:SetDuration(0.5)
	button.Fader.Anim:SetSmoothing('IN_OUT')
	
	button:HookScript('OnEnter', function() OnEnterAura(auras:GetParent(), button) end)
    button:HookScript('OnLeave', function() OnLeaveAura(auras:GetParent()) end)
end

-- update auras icon
local PostUpdateIcon = function(auras, unit, button, index, _, duration, expiration, debuffType)
	-- set auras type
	local Frenzy = LCD:UnitAura(unit, index, 'HELPFUL')
	button.isFriend = UnitIsFriend('player', unit)
	button.isEnemyBuff = not button.isFriend and not button.isDebuff
	button.canDispel = button.isFriend and DispelList[debuffType]
	button.canPurge = button.isEnemyBuff and debuffType == 'Magic'
	button.isFrenzy = button.isEnemyBuff and filter.Frenzy[Frenzy]
	-- set auras timer
	if (duration and duration > 0) and (button.isFriend or button.isPlayer or button.isEnemyBuff or button.canPurge) then
		button.expiration = expiration - GetTime()
		button:SetScript('OnUpdate', CreateAuraTimer)
	else
		button.time:SetText()
		button.expiration = huge
	end
	-- set auras classification
	if (button.isDebuff) then
		if (not button.isFriend and not button.isPlayer) then
			button.icon:SetDesaturated(true)
			button.overlay:SetVertexColor(unpack(cfg.maincolor))
		else
			local color = _G.DebuffTypeColor[debuffType or 'none']
			button.icon:SetDesaturated(false)
			button.overlay:SetVertexColor(color.r*0.9, color.g*0.9, color.b*0.9)
		end
		if (button.canDispel) then
			LCG.PixelGlow_Start(button, cfg.countcolor, 2, 0.5, 15, nil, 1, 1)
		else
			LCG.PixelGlow_Stop(button)
		end
	else
		if (button.isFriend and button.isPlayer) then
			button.icon:SetDesaturated(false)
		else
			button.icon:SetDesaturated(true)
		end
		if (button.canPurge or button.isFrenzy) then
			LCG.PixelGlow_Start(button, cfg.countcolor, 4, nil, 5, nil, 1, 1)
			button.icon:SetDesaturated(false)
		else
			LCG.PixelGlow_Stop(button)
		end
	end
	-- right click cancel buffs not incombat
	button:SetScript('OnClick', function(self, Button)
		if (Button == 'RightButton') then
			OnLeaveAura(auras:GetParent())
			if (not InCombatLockdown()) then
				CancelUnitBuff('player', index)
			end
		end
	end)
end

-- sort auras by time remaining
local SortAuras = function(a, b)
	if (a:IsShown() and b:IsShown()) then
		return a.expiration < b.expiration
	else
		return a:IsShown()
	end
end

local PreSetPosition = function(auras)
	tsort(auras, SortAuras)
	return 1, auras.visibleBuffs or auras.visibleDebuffs
end

-- relative anchor point for the target debuffs
local UpdateDebuffsPosition = function(auras)
	local Auras = auras.__owner
	local Buffs = Auras.Buffs
	local Debuffs = Auras.Debuffs
	local VisibleBuffs = auras.visibleBuffs

	Debuffs:ClearAllPoints()
	
	if (VisibleBuffs == 0) then
		Debuffs:SetAllPoints(Buffs)
	elseif (VisibleBuffs < 9) then
		Debuffs:SetPoint('BOTTOMLEFT', Buffs, 0, Buffs.size + Buffs.spacing)
	else
		Debuffs:SetPoint('BOTTOMLEFT', Buffs, 0, Buffs.size * 2 + Buffs.spacing * 2)
	end
end

-- create buffs
local CreateBuffs = function(self)
    local Buffs = CreateFrame('Frame', nil, self)
	Buffs.PreSetPosition = PreSetPosition
	Buffs.PostCreateIcon = PostCreateIcon
	Buffs.PostUpdateIcon = PostUpdateIcon
	Buffs.forceShow = cfg.forceShow

	if (self.unit == 'target') then
        Buffs.PostUpdate = UpdateDebuffsPosition
    end

	self.Buffs = Buffs
end

-- create debuffs
local CreateDebuffs = function(self)
	local Debuffs = CreateFrame('Frame', nil, self)
	Debuffs.PreSetPosition = PreSetPosition
	Debuffs.PostCreateIcon = PostCreateIcon
	Debuffs.PostUpdateIcon = PostUpdateIcon
	Debuffs.forceShow = cfg.forceShow
	self.Debuffs = Debuffs
end

-- mouseover zoom auras
local HighlightAura = function(self)
	local HighlightAura = CreateFrame('Frame', nil, self)
    HighlightAura.button = HighlightAura:CreateTexture(nil, 'ARTWORK')
    HighlightAura.button:SetPoint('CENTER')
    HighlightAura.overlay = HighlightAura:CreateTexture(nil, 'OVERLAY')
    HighlightAura.overlay:SetTexture(cfg.Auratex)
    HighlightAura.overlay:SetPoint('CENTER')
    
    self.HighlightAura = HighlightAura
end

--------------------------------
-- Shared stuff for all units --
--------------------------------
local Shared = function(self, unit)
	-- register click
	self:RegisterForClicks('AnyUp')
    self:SetScript('OnEnter', _G.UnitFrame_OnEnter)
	self:SetScript('OnLeave', _G.UnitFrame_OnLeave)
	
    -- health Bar
    local Health = CreateFrame('StatusBar', nil, self)
	Health:SetPoint('TOP')
    Health:SetPoint('LEFT')
	Health:SetPoint('RIGHT')
	Health:SetStatusBarTexture(cfg.Normal)
    Health:SetStatusBarColor(unpack(cfg.maincolor))
    Health:SetFrameLevel(3)
	
    -- health bar background
    local Background = Health:CreateTexture(nil, 'BACKGROUND')
    Background:SetAllPoints(Health)
    Background:SetTexture(cfg.NormBG)
	Background:SetVertexColor(unpack(cfg.hpBGcolor))
    Background:SetAlpha(0.6)
	
    -- health bar border
    local Border = CreateFrame('Frame', nil, Health, 'BackdropTemplate')
    Border:SetPoint('TOPLEFT', Health, -3, 3)
    Border:SetPoint('BOTTOMRIGHT', Health, 3, -3)
    Border:SetBackdrop(cfg.frameBD)
    Border:SetBackdropBorderColor(0, 0, 0, 1)
    Border:SetFrameLevel(3)
	
	self.Health = Health
	Health.Smooth = true
	Health.PostUpdate = PostUpdateHealth

	-- health arrow
	local Arrow = Health:CreateTexture(nil, 'OVERLAY')
	Arrow:SetTexture(cfg.Arrow)
	Arrow:SetPoint('BOTTOM', Health:GetStatusBarTexture(), 'TOPRIGHT')
	Arrow:SetSize(12, 12)
	self.Arrow = Arrow

    -- power bar
    local Power = CreateFrame('StatusBar', nil, self)
    Power:SetStatusBarTexture(cfg.Normal)
	Power:SetPoint('TOP', Health, 'BOTTOM', 0, -5)
    Power:SetFrameLevel(4)

    -- power bar background
    local Background = Power:CreateTexture(nil, 'BACKGROUND')
    Background:SetAllPoints(Power)
    Background:SetTexture(cfg.NormBG)
    Background:SetVertexColor(unpack(cfg.pbBGcolor))
    Background:SetAlpha(0.6)
    
    -- power bar border
    local Border = CreateFrame('Frame', nil, Power, 'BackdropTemplate')
    Border:SetPoint('TOPLEFT', Power, -3, 3)
    Border:SetPoint('BOTTOMRIGHT', Power, 3, -3)
    Border:SetBackdrop(cfg.frameBD)
    Border:SetBackdropBorderColor(0, 0, 0, 1)
	Border:SetFrameLevel(4)
	
	Power.Smooth = true
	Power.colorPower = true
	Power.frequentUpdates = true

	if (self.unit ~= 'raid') then
		Power.PostUpdate = PostUpdatePower
	else
		Power.PostUpdate = PostUpdatePowerRaid
	end

	self.Power = Power
	self.colors = Powercolor
	
	-- status strings
    local status = Health:CreateFontString(nil, 'OVERLAY')
    status:SetFont(cfg.PlugFont, cfg.NormFS, cfg.FontF)
	self:Tag(status, '[difficulty][level][shortclassification]')
	self.status = status

    local name = Health:CreateFontString(nil, 'OVERLAY')
    name:SetFont(cfg.NameFont, cfg.NormFS, cfg.FontF)
	name:SetWordWrap(false)
    self:Tag(name, '[namecolor][name]')
	self.name = name
    
	local perhp = Health:CreateFontString(nil, 'OVERLAY')
	perhp:SetFont(cfg.PlugFont, cfg.NormFS, cfg.FontF)
	self.perhp = perhp

	local hpp = Health:CreateFontString(nil, 'OVERLAY')
	hpp:SetFont(cfg.AuraFont, cfg.LargeFS, cfg.FontF)
	self:Tag(hpp, '[hpp]')
	self.hpp = hpp
	
	-- cur/max values
	local hp = Health:CreateFontString(nil, 'OVERLAY')
	hp:SetFont(cfg.NumbFont, cfg.NormFS, cfg.FontF)
	self:Tag(hp, '[hp]')
	hp:SetPoint('CENTER', Health)
	hp:SetAlpha(0)
	self.hp = hp

	local pp = Power:CreateFontString(nil, 'OVERLAY')
	pp:SetFont(cfg.NumbFont, cfg.SmallFS, cfg.FontF)
	self:Tag(pp, '[pp]')
	pp:SetPoint('CENTER', Power)
	pp:SetAlpha(0)
	self.pp = pp

	-- cur values
	local curhp = Health:CreateFontString(nil, 'OVERLAY')
	curhp:SetFont(cfg.CurrFont, cfg.SmallFS, cfg.FontF)
	self:Tag(curhp, '[curhp]')
	self.curhp = curhp

	local curpp = Power:CreateFontString(nil, 'OVERLAY')
	curpp:SetFont(cfg.CurrFont, cfg.SmallFS, cfg.FontF)
	self:Tag(curpp, '[curpp]')
	self.curpp = curpp
end

-- castbar
local Castbar = function(self)
    local cb = CreateFrame('StatusBar', nil, self, 'BackdropTemplate')
	cb:SetStatusBarTexture(cfg.Castbar)
	cb:SetAllPoints(self.Health)
	cb:SetFrameLevel(self.Health:GetFrameLevel() + 1)

	-- Castbar color
    cb.CastingColor = {0, 1, 0, 0.5}
    cb.CompleteColor = {0, 1, 0, 1}
    cb.FailColor = {1, 0.2, 0.2, 1}
    cb.ChannelingColor = {0.1, 0.6, 1, 0.6}
    
	cb.Text = cb:CreateFontString(nil, 'OVERLAY')
    cb.Text:SetFont(cfg.NameFont, cfg.NormFS, cfg.FontF)
	cb.Text:SetPoint('LEFT', cb, 4, -1)
	cb.Text:SetJustifyH('LEFT')
    
	cb.Time = cb:CreateFontString(nil, 'OVERLAY')
    cb.Time:SetFont(cfg.PlugFont, cfg.NormFS, cfg.FontF)
	cb.Time:SetPoint('RIGHT', cb, 0, -1)
	cb.Time:SetJustifyH('RIGHT')
    
    cb.Icon = cb:CreateTexture(nil, 'OVERLAY')
    cb.Icon:SetPoint('RIGHT', cb, 'LEFT', -4, 0)
    cb.Icon:SetTexCoord(.08, .92, .08, .92)
    
    cb.IconGlow = CreateFrame('Frame', nil, cb, 'BackdropTemplate')
    cb.IconGlow:SetPoint('TOPLEFT', cb.Icon, -1, 1)
    cb.IconGlow:SetPoint('BOTTOMRIGHT', cb.Icon, 1, -1)
	cb.IconGlow:SetBackdrop(cfg.pixelBD)
	cb.IconGlow:SetBackdropColor(0, 0, 0, 0)
	cb.IconGlow:SetBackdropBorderColor(0, 0, 0, 1)
    
    cb.Spark = cb:CreateTexture(nil, 'OVERLAY')
    cb.Spark:SetBlendMode('ADD')
    cb.Spark:SetSize(16, 48)
	cb.Spark:SetPoint('CENTER', cb:GetStatusBarTexture(), 'RIGHT')
	
	cb.OnUpdate = castbar.OnCastbarUpdate
    cb.PostCastStart = castbar.PostCastStart
    cb.PostCastStop = castbar.PostCastStop
    cb.PostCastFail = castbar.PostCastFail

	self.Castbar = cb
end

-- mouseover highlight
local movHlight = function(self)
    local movHlight = self.Health:CreateTexture(nil, 'ARTWORK')
    movHlight:SetPoint('TOPLEFT', self.Health, -30, 4)
    movHlight:SetPoint('TOPRIGHT', self.Health, 30, 4)
    movHlight:SetHeight(3)
    movHlight:SetTexture(cfg.movH)
    movHlight:SetVertexColor(1, 1, 1, 1)
    movHlight:SetTexCoord(0, 1, 1, 0)
	movHlight:Hide()
	movHlight:SetBlendMode('ADD')

    self.movHlight = movHlight
end

-- threat highlight
local threatHlight = function(self)
	local Thrt = self.Health:CreateTexture(nil, 'OVERLAY')
	Thrt:SetPoint('TOPLEFT', self.Health, -40, 4)
	Thrt:SetPoint('TOPRIGHT', self.Health, 40, 4)
	Thrt:SetHeight(2)
	Thrt:SetTexture(cfg.thdH)
	Thrt:Hide()
	Thrt:SetBlendMode('ADD')

	self.ThreatIndicator = Thrt
end

local threatGlow = function(self)
	local Thrt = CreateFrame('Frame', nil, self, 'BackdropTemplate')
	Thrt:SetPoint('TOPLEFT', self.Health, -5, 5)
    Thrt:SetPoint('BOTTOMRIGHT', self.Health, 5, -5)
	Thrt:SetBackdrop(cfg.frameBD)
	Thrt:SetFrameLevel(self.Health:GetFrameLevel() + 1)
	Thrt:Hide()

	self.ThreatIndicator = Thrt
end

-- debuff highlight
local dHlight = function(self)
	local dHlight = self.Health:CreateTexture(nil, 'OVERLAY')
	dHlight:SetPoint('BOTTOMLEFT', self.Health, -40, -4)
	dHlight:SetPoint('BOTTOMRIGHT', self.Health, 40, -4)
	dHlight:SetHeight(2)
	dHlight:SetTexture(cfg.thdH)
	dHlight:SetBlendMode('ADD')
	dHlight:Hide()

	self.DebuffHighlight = dHlight
end

local rdHlight = function(self)
	local rdHlight = CreateFrame('Frame', nil, self, 'BackdropTemplate')
	rdHlight:SetPoint('TOPLEFT', self.Health, -4, 4)
    rdHlight:SetPoint('BOTTOMRIGHT', self.Health, 4, -4)
	rdHlight:SetBackdrop(cfg.frameBD)
	rdHlight:SetFrameLevel(self.Health:GetFrameLevel() + 2)
	rdHlight:Hide()

	self.DebuffHighlightBorder = rdHlight
end

-- target highlight
local updatetargetGlow = function(self, event, unit)
	if UnitIsUnit(self.unit, 'target') then
		self.targetGlow:Show()
		if (not self.tgFader:IsPlaying()) then
			self.tgFader:Play()
		end
	else
		self.targetGlow:Hide()
		if (self.tgFader:IsPlaying()) then
			self.tgFader:Stop()
		end
	end
end

local targetGlow = function(self)
	local targetGlow = CreateFrame('Frame', nil, self, 'BackdropTemplate')
	targetGlow:SetPoint('TOPLEFT', self.Health, -4, 4)
    targetGlow:SetPoint('BOTTOMRIGHT', self.Health, 4, -4)
	targetGlow:SetBackdrop(cfg.frameBD)
	targetGlow:SetBackdropBorderColor(1, 1, 0)
	targetGlow:SetFrameLevel(self.Health:GetFrameLevel() + 3)
	targetGlow:Hide()
	self.targetGlow = targetGlow

	local tgFader = targetGlow:CreateAnimationGroup()
	tgFader:SetLooping('BOUNCE')

	tgFader.Anim = tgFader:CreateAnimation('Alpha')
	tgFader.Anim:SetFromAlpha(1)
	tgFader.Anim:SetToAlpha(0.5)
	tgFader.Anim:SetDuration(0.5)
	tgFader.Anim:SetSmoothing('IN_OUT')
	self.tgFader = tgFader

	self:RegisterEvent('PLAYER_TARGET_CHANGED', updatetargetGlow, true)
end

-- fader
local Fader = function(self)
    self.Fader = cfg.Fader
    self.FadeIn = cfg.Fader.FadeIn
    self.FadeOut = cfg.Fader.FadeOut
    self.MinAlpha = cfg.Fader.MinAlpha
    self.MaxAlpha = cfg.Fader.MaxAlpha
end

-- Portrait
local Portrait = function(self)
	Portrait = CreateFrame('PlayerModel', nil, self)
	Portrait:SetAllPoints(self.Health)
	Portrait:SetFrameLevel(self.Health:GetFrameLevel())
	Portrait:SetAlpha(cfg.pAlpha)

	self.Portrait = Portrait
end

-- spell range
local SpellRange = function(self)
	self.Range = {
		insideAlpha = 1,
		outsideAlpha = 0.6
	}
end

-- heal prediction
local HealPrediction = function(self)
	local health = self.Health

	local myBar = CreateFrame('StatusBar', nil, health)
	myBar:SetPoint('TOPLEFT', health:GetStatusBarTexture(), 'TOPRIGHT')
	myBar:SetPoint('BOTTOMLEFT', health:GetStatusBarTexture(), 'BOTTOMRIGHT')
	myBar:SetWidth(self:GetWidth())
	myBar:SetStatusBarTexture(cfg.Healbar)
	myBar:SetStatusBarColor(unpack(cfg.SelfColor))
	myBar:SetAlpha(0.6)
	myBar:SetFrameLevel(health:GetFrameLevel())
	myBar.Smooth = true
	
	local otherBar = CreateFrame('StatusBar', nil, health)
	otherBar:SetPoint('TOPLEFT', myBar:GetStatusBarTexture(), 'TOPRIGHT')
	otherBar:SetPoint('BOTTOMLEFT', myBar:GetStatusBarTexture(), 'BOTTOMRIGHT')
	otherBar:SetWidth(self:GetWidth())
	otherBar:SetStatusBarTexture(cfg.Healbar)
	otherBar:SetStatusBarColor(unpack(cfg.OtherColor))
	otherBar:SetAlpha(0.6)
	otherBar:SetFrameLevel(health:GetFrameLevel())
	otherBar.Smooth = true
	
	self.HealthPrediction = {
		myBar = myBar,
		otherBar = otherBar,
		maxOverflow = 1.0,
	}
end

-------------------------
-- Unit specific stuff --
-------------------------
local UnitSpecific = {
	player = function(self, ...)
		Shared(self, ...)
		Fader(self)
		Castbar(self)
		dHlight(self)
		movHlight(self)
		Portrait(self)
		threatHlight(self)
		HighlightAura(self)
		
		-- Unit Frame Size
		self:SetSize(cfg.UFS.Prime.w, cfg.UFS.Prime.h)
		self.Health:SetHeight(cfg.UFS.Prime.hh)
		self.Power:SetSize(cfg.UFS.Prime.w, cfg.UFS.Prime.ph)
		self.Castbar.Text:SetWidth(cfg.UFS.Prime.w * 0.7)
		self.Castbar.Icon:SetSize(cfg.UFS.Prime.hh, cfg.UFS.Prime.hh)

		-- Switch element
		self:HookScript('OnEnter', function()
			self.curhp:SetAlpha(0)
			self.curpp:SetAlpha(0)
			self.movHlight:Show()
			UIFrameFadeOut(self.hp, 0.5, self.hp:GetAlpha(), 1)
			UIFrameFadeOut(self.pp, 0.5, self.pp:GetAlpha(), 1)
			UIFrameFadeOut(self.Castbar, 0, self.Castbar:GetAlpha(), 0)
			UIFrameFadeOut(self.element, cfg.delay, self.element:GetAlpha(), 0)
			UIFrameFadeOut(self.Portrait, cfg.delay, self.Portrait:GetAlpha(), 0)
		end)
		
		self:HookScript('OnLeave', function()
			self.curhp:SetAlpha(1)
			self.curpp:SetAlpha(1)
			self.movHlight:Hide()
			UIFrameFadeIn(self.hp, 0.5, self.hp:GetAlpha(), 0)
			UIFrameFadeIn(self.pp, 0.5, self.pp:GetAlpha(), 0)
			UIFrameFadeIn(self.Castbar, cfg.delay, self.Castbar:GetAlpha(), 1)
			UIFrameFadeIn(self.element, cfg.delay, self.element:GetAlpha(), 1)
			UIFrameFadeIn(self.Portrait, cfg.delay, self.Portrait:GetAlpha(), cfg.pAlpha)
		end)

		-- Auras
		CreateDebuffs(self)
		self.Debuffs:SetPoint('TOPLEFT', self.Power, 'BOTTOMLEFT', 2, -15)
		self.Debuffs.num = 6
		self.Debuffs.size = cfg.auraSize
		self.Debuffs.spacing = 5
		self.Debuffs:SetSize(self.Debuffs.num * (self.Debuffs.size + self.Debuffs.spacing), self.Debuffs.size + self.Debuffs.spacing)

		-- Aurabar
		if cfg.Aurabar then
			local Aurabar = aurabar.CreateBar(self)
    		Aurabar:SetPoint('BOTTOMLEFT', self, 'TOPLEFT', 0, 75)
			Aurabar.CustomFilter = aurabar.Filter
			self.Auras = Aurabar
		end

		-- EnergyManaRegen
		if cfg.EMR then
			local EMR = CreateFrame('StatusBar', nil, self)
			EMR:SetAllPoints(self.Power)
			EMR:SetFrameLevel(self.Power:GetFrameLevel() + 3)
			EMR.Spark = EMR:CreateTexture(nil, 'OVERLAY')
			self.EnergyManaRegen = EMR
		end

		-- Totem icons
		if cfg.Totems then
			local Totems = {}
			for index = 1, 4 do
				local Totem = CreateFrame('Button', nil, self)
				Totem:SetSize(24, 24)
				Totem:SetPoint('BOTTOMLEFT', self, 'TOPLEFT', (index - 1) * Totem:GetWidth() * 1.2, 15)
	
				local Icon = Totem:CreateTexture(nil, 'OVERLAY')
				Icon:SetAllPoints()
	
				local Cooldown = CreateFrame('Cooldown', nil, Totem, 'CooldownFrameTemplate')
				Cooldown:SetAllPoints()
	
				Totem.Icon = Icon
				Totem.Cooldown = Cooldown
    			Totem.Cooldown.noCooldownCount = cfg.hideCDc
	
				Totems[index] = Totem
			end
			self.Totems = Totems
		end
		
		-- xp/rep
		if cfg.showXpRep then
			local XpRep = CreateFrame('Frame', nil, self)
			XpRep:SetAllPoints(self)
		
			local xprep = XpRep:CreateFontString(nil, 'OVERLAY')
			xprep:SetFont(cfg.NameFont, cfg.SmallFS, cfg.FontF)
			xprep:SetPoint('TOP', self.Power, 'BOTTOM', 0, -5)
			self:Tag(xprep, '[Exp] [Rep]')
			xprep:SetAlpha(0)
			self.xprep = xprep
			
			self:HookScript('OnEnter', function()
				UIFrameFadeIn(xprep, cfg.delay, xprep:GetAlpha(), 1)
			end)
			
			self:HookScript('OnLeave', function()
				UIFrameFadeOut(xprep, cfg.delay, xprep:GetAlpha(), 0)
			end)
		end

		-- HealComm
		HealPrediction(self)
		
		-- Element
		local element = CreateFrame('Frame', nil, self)
		element:SetAllPoints(self.Health)
		element:SetFrameLevel(self.Health:GetFrameLevel())
		self.element = element
		
		CIi = element:CreateTexture(nil, 'OVERLAY')
		CIi:SetSize(26, 24)
		CIi:SetPoint('LEFT', self.Health, 2, 1)
		CIi:SetTexture(cfg.Cit)
		self.CombatIndicator = CIi

		LIi = element:CreateTexture(nil, 'OVERLAY')
		LIi:SetSize(22, 22)
		LIi:SetPoint('LEFT', CIi, 'RIGHT', -1, -1)
		LIi:SetTexture(cfg.Lit)
		self.LeaderIndicator = LIi

		MLi = element:CreateTexture(nil, 'OVERLAY')
		MLi:SetSize(20, 20)
		MLi:SetPoint('LEFT', LIi, 'RIGHT', 0, 1)
		MLi:SetTexture(cfg.MLi)
		self.MasterLooterIndicator = MLi

		Afk = element:CreateTexture(nil, 'OVERLAY')
		Afk:SetSize(16, 16)
		Afk:SetPoint('TOPRIGHT', self.Health, 2, 1)
		Afk:SetTexture(cfg.Afk)
		self.AfkIndicator = Afk

		Dnd = element:CreateTexture(nil, 'OVERLAY')
		Dnd:SetSize(16, 16)
		Dnd:SetPoint('TOPRIGHT', self.Health, 2, 1)
		Dnd:SetTexture(cfg.Dnd)
		self.DndIndicator = Dnd

		FDi = element:CreateTexture(nil, 'OVERLAY')
    	FDi:SetSize(20, 20)
    	FDi:SetPoint('CENTER', self.Health)
		self.FeignIndicator = FDi

		Res = element:CreateTexture(nil, 'OVERLAY')
    	Res:SetSize(20, 20)
    	Res:SetPoint('CENTER', self.Health)
		self.ResurrectIndicator = Res

		RSi = element:CreateTexture(nil, 'OVERLAY')
		RSi:SetSize(16, 16)
		RSi:SetPoint('TOPRIGHT', self.Health, 9, 9)
		RSi:SetTexture(cfg.Rit)
		self.RestingIndicator = RSi
		
		RTi = self:CreateTexture(nil, 'OVERLAY')
		RTi:SetSize(22, 22)
		RTi:SetPoint('LEFT', self.Health, 'RIGHT', 4, 0)
		RTi:SetTexture(cfg.RTi)
		self.RaidTargetIndicator = RTi
	end,
	
	pet = function(self, ...)
		Shared(self, ...)
		Fader(self)
		Castbar(self)
		dHlight(self)
		movHlight(self)
		Portrait(self)
		threatHlight(self)
		
		-- Unit Frame Size
		self:SetSize(cfg.UFS.Pet.w, cfg.UFS.Pet.h)
		self.Health:SetHeight(cfg.UFS.Pet.hh)
		self.Power:SetSize(cfg.UFS.Pet.w * 0.6, cfg.UFS.Pet.ph)
		self.Power:SetPoint('TOP', self.Health, 'BOTTOM')
		self.Castbar.Text:SetWidth(cfg.UFS.Pet.w * 0.6)
		self.Castbar.Icon:SetSize(cfg.UFS.Pet.hh + 2, cfg.UFS.Pet.hh + 2)
		
		-- Font strings
		self.name:SetPoint('BOTTOMLEFT', self.Health, 'TOPLEFT', 0, 2)
		self.hpp:SetPoint('BOTTOMRIGHT', self.Health, 'TOPRIGHT', 3, 2)
		
		-- Switch element
		self:HookScript('OnEnter', function()
			self.movHlight:Show()
			UIFrameFadeOut(self.hp, 0.5, self.hp:GetAlpha(), 1)
			UIFrameFadeOut(self.Castbar, 0, self.Castbar:GetAlpha(), 0)
			UIFrameFadeOut(self.element, cfg.delay, self.element:GetAlpha(), 0)
			UIFrameFadeOut(self.Portrait, cfg.delay, self.Portrait:GetAlpha(), 0)
		end)
		
		self:HookScript('OnLeave', function()
			self.movHlight:Hide()
			UIFrameFadeIn(self.hp, 0.5, self.hp:GetAlpha(), 0)
			UIFrameFadeIn(self.Castbar, cfg.delay, self.Castbar:GetAlpha(), 1)
			UIFrameFadeIn(self.element, cfg.delay, self.element:GetAlpha(), 1)
			UIFrameFadeIn(self.Portrait, cfg.delay, self.Portrait:GetAlpha(), cfg.pAlpha)
		end)
		
		-- HealComm
		HealPrediction(self)
		
		-- Element
		local element = CreateFrame('Frame', nil, self)
		element:SetAllPoints(self.Health)
		element:SetFrameLevel(self.Health:GetFrameLevel())
		self.element = element
		
		CIi = element:CreateTexture(nil, 'OVERLAY')
		CIi:SetSize(20, 18)
		CIi:SetPoint('LEFT', self.Health, 2, 1)
		CIi:SetTexture(cfg.Cit)
		self.CombatIndicator = CIi

		PHi = CreateFrame('Frame', nil, self)
		PHi:SetPoint('RIGHT', self.Health, 'LEFT', -4, 0)
		PHi:SetSize(18, 18)
    	self.HappinessIndicator = PHi
	
		RTi = self:CreateTexture(nil, 'OVERLAY')
		RTi:SetSize(18, 18)
		RTi:SetPoint('LEFT', self.Health, 'RIGHT', 2, 0)
		RTi:SetTexture(cfg.RTi)
		self.RaidTargetIndicator = RTi
	end,
	
	target = function(self, ...)
		Shared(self, ...)
		Castbar(self)
		dHlight(self)
		movHlight(self)
		Portrait(self)
		threatHlight(self)
		HighlightAura(self)

		-- Unit Frame Size
		self:SetSize(cfg.UFS.Prime.w, cfg.UFS.Prime.h)
		self.Health:SetHeight(cfg.UFS.Prime.hh)
		self.Power:SetSize(cfg.UFS.Prime.w, cfg.UFS.Prime.ph)
		self.Castbar.Text:SetWidth(cfg.UFS.Prime.w * 0.7)
		self.Castbar.Icon:SetSize(cfg.UFS.Prime.hh, cfg.UFS.Prime.hh)

		-- Status strings
		self.status:SetPoint('BOTTOMLEFT', self.Health, 'TOPLEFT', 0, 2)
		self.name:SetWidth(self:GetWidth() * 0.7)
		self.name:SetPoint('LEFT', self.status, 'RIGHT', 0, 0)
		self.name:SetJustifyH('LEFT')
		self.name:SetWordWrap(false)
		self.perhp:SetPoint('BOTTOMRIGHT', self.Health, 'TOPRIGHT', 3, 2)

		-- Switch element
		self:HookScript('OnEnter', function()
			self.curhp:SetAlpha(0)
			self.curpp:SetAlpha(0)
			self.movHlight:Show()
			UIFrameFadeOut(self.hp, 0.5, self.hp:GetAlpha(), 1)
			UIFrameFadeOut(self.pp, 0.5, self.pp:GetAlpha(), 1)
			UIFrameFadeOut(self.Castbar, 0, self.Castbar:GetAlpha(), 0)
			UIFrameFadeOut(self.element, cfg.delay, self.element:GetAlpha(), 0)
			UIFrameFadeOut(self.Portrait, cfg.delay, self.Portrait:GetAlpha(), 0)
		end)
		
		self:HookScript('OnLeave', function()
			self.curhp:SetAlpha(1)
			self.curpp:SetAlpha(1)
			self.movHlight:Hide()
			UIFrameFadeIn(self.hp, 0.5, self.hp:GetAlpha(), 0)
			UIFrameFadeIn(self.pp, 0.5, self.pp:GetAlpha(), 0)
			UIFrameFadeIn(self.Castbar, cfg.delay, self.Castbar:GetAlpha(), 1)
			UIFrameFadeIn(self.element, cfg.delay, self.element:GetAlpha(), 1)
			UIFrameFadeIn(self.Portrait, cfg.delay, self.Portrait:GetAlpha(), cfg.pAlpha)
		end)

		-- Auras
		CreateBuffs(self)
		self.Buffs:SetPoint('BOTTOMLEFT', self, 'TOPLEFT', 0, 24)
		self.Buffs.num = 16
		self.Buffs.size = cfg.auraSize
		self.Buffs.spacing = 4
		self.Buffs:SetSize((self.Buffs.num / 2) * (self.Buffs.size + self.Buffs.spacing), self.Buffs.size + self.Buffs.spacing)
		
		CreateDebuffs(self)
		self.Debuffs.num = 16
		self.Debuffs.size = cfg.auraSize
		self.Debuffs.spacing = 4
		self.Debuffs:SetSize((self.Debuffs.num / 2) * (self.Debuffs.size + self.Debuffs.spacing), self.Debuffs.size * 2 + self.Debuffs.spacing)

		-- Aurabar
		if cfg.Aurabar then
			local Aurabar = aurabar.CreateBar(self)
    		Aurabar:SetPoint('BOTTOMLEFT', self, 'TOPLEFT', 0, 145)
			Aurabar.CustomFilter = aurabar.Filter
			self.Auras = Aurabar
		end

		-- Combo Points
		local ComboPoints = {}
    	for i = 1, 5 do
			local Bar = CreateFrame('StatusBar', nil, self)
			Bar:SetSize(25, 25)
			Bar:SetStatusBarTexture(cfg.CPt)
			Bar:SetPoint('TOPLEFT', self.Power, 'BOTTOMLEFT', (i - 1) * Bar:GetWidth() * 1.5, 0)
        	ComboPoints[i] = Bar
		end
		self.ComboPoints = ComboPoints

		-- HealComm
		HealPrediction(self)

		-- Element
		local element = CreateFrame('Frame', nil, self)
		element:SetAllPoints(self.Health)
		element:SetFrameLevel(self.Health:GetFrameLevel())
		self.element = element

		CIi = element:CreateTexture(nil, 'OVERLAY')
		CIi:SetSize(24, 22)
		CIi:SetPoint('LEFT', self.Health, 2, 1)
		CIi:SetTexture(cfg.Cit)
		self.CombatIndicator = CIi

		LIi = element:CreateTexture(nil, 'OVERLAY')
		LIi:SetSize(20, 20)
		LIi:SetPoint('LEFT', CIi, 'RIGHT', -1, -1)
		LIi:SetTexture(cfg.Lit)
		self.LeaderIndicator = LIi

		Afk = element:CreateTexture(nil, 'OVERLAY')
		Afk:SetSize(16, 16)
		Afk:SetPoint('TOPRIGHT', self.Health, 2, 1)
		Afk:SetTexture(cfg.Afk)
		self.AfkIndicator = Afk
		
		Dnd = element:CreateTexture(nil, 'OVERLAY')
		Dnd:SetSize(16, 16)
		Dnd:SetPoint('TOPRIGHT', self.Health, 2, 1)
		Dnd:SetTexture(cfg.Dnd)
		self.DndIndicator = Dnd

		FDi = element:CreateTexture(nil, 'OVERLAY')
    	FDi:SetSize(20, 20)
    	FDi:SetPoint('CENTER', self.Health)
		self.FeignIndicator = FDi

		Off = element:CreateTexture(nil, 'OVERLAY')
    	Off:SetSize(30, 30)
    	Off:SetPoint('CENTER', self.Health)
    	self.OfflineIndicator = Off

		Res = element:CreateTexture(nil, 'OVERLAY')
    	Res:SetSize(20, 20)
    	Res:SetPoint('CENTER', self.Health)
		self.ResurrectIndicator = Res

		RTi = self:CreateTexture(nil, 'OVERLAY')
		RTi:SetSize(22, 22)
		RTi:SetPoint('LEFT', self.Health, 'RIGHT', 4, 0)
		RTi:SetTexture(cfg.RTi)
		self.RaidTargetIndicator = RTi
	end,

	targettarget = function(self, ...)
		Shared(self, ...)
		dHlight(self)
		movHlight(self)
		Portrait(self)
		
		-- Unit Frame Size
		self:SetSize(cfg.UFS.TT.w, cfg.UFS.TT.h)
		self.Health:SetHeight(cfg.UFS.TT.hh)
		self.Power:SetSize(cfg.UFS.TT.w * 0.6, cfg.UFS.TT.ph)
		self.Power:SetPoint('TOP', self.Health, 'BOTTOM')

		-- Font strings
		self.name:SetWidth(self:GetWidth() * 0.8)
		self.name:SetPoint('LEFT', self.Health, 2, 0)
		self.name:SetJustifyH('LEFT')
		self.name:SetWordWrap(false)
		self.hpp:SetPoint('RIGHT', self.Health, 0, 2)

		-- Switch element
		self:HookScript('OnEnter', function()
			self.movHlight:Show()
			UIFrameFadeOut(self.hp, 0.5, self.hp:GetAlpha(), 1)
			UIFrameFadeOut(self.hpp, 0.5, self.hpp:GetAlpha(), 0)
			UIFrameFadeOut(self.name, 0.5, self.name:GetAlpha(), 0)
			UIFrameFadeOut(self.Portrait, cfg.delay, self.Portrait:GetAlpha(), 0)
		end)
		
		self:HookScript('OnLeave', function()
			self.movHlight:Hide()
			UIFrameFadeIn(self.hp, 0.5, self.hp:GetAlpha(), 0)
			UIFrameFadeIn(self.hpp, 0.5, self.hpp:GetAlpha(), 1)
			UIFrameFadeIn(self.name, 0.5, self.name:GetAlpha(), 1)
			UIFrameFadeIn(self.Portrait, cfg.delay, self.Portrait:GetAlpha(), cfg.pAlpha)
		end)
		
		-- HealComm
		HealPrediction(self)

		-- Element
		RTi = self:CreateTexture(nil, 'OVERLAY')
		RTi:SetSize(18, 18)
		RTi:SetPoint('BOTTOM', self.Health, 'TOP', 0, -1)
		RTi:SetTexture(cfg.RTi)
		self.RaidTargetIndicator = RTi
	end,

	tank = function(self, ...)
		Shared(self, ...)
		dHlight(self)
		movHlight(self)
		Portrait(self)
		SpellRange(self)
		threatHlight(self)
		
		-- Unit Frame Size
		self.Health:SetHeight(cfg.UFS.Tank.hh)
		self.Power:SetSize(cfg.UFS.Tank.w * 0.6, cfg.UFS.Tank.ph)
		self.Power:SetPoint('TOP', self.Health, 'BOTTOM')

		-- Font strings
		self.name:SetWidth(self:GetWidth() * 0.8)
		self.name:SetPoint('BOTTOMLEFT', self.Health, 'TOPLEFT', 0, 2)
		self.name:SetJustifyH('LEFT')
		self.name:SetWordWrap(false)
		self.hpp:SetPoint('BOTTOMRIGHT', self.Health, 'TOPRIGHT', 3, 2)

		-- Switch element
		self:HookScript('OnEnter', function()
			self.movHlight:Show()
			UIFrameFadeOut(self.hp, 0.5, self.hp:GetAlpha(), 1)
			UIFrameFadeOut(self.Portrait, cfg.delay, self.Portrait:GetAlpha(), 0)
		end)
		
		self:HookScript('OnLeave', function()
			self.movHlight:Hide()
			UIFrameFadeIn(self.hp, 0.5, self.hp:GetAlpha(), 0)
			UIFrameFadeIn(self.Portrait, cfg.delay, self.Portrait:GetAlpha(), cfg.pAlpha)
		end)

		-- HealComm
		HealPrediction(self)

		-- DebuffHighlightFilter
		self.DebuffHighlightFilter = true
		
		-- Element
		RTi = self:CreateTexture(nil, 'OVERLAY')
		RTi:SetSize(18, 18)
		RTi:SetPoint('RIGHT', self.Health, 'LEFT', -4, 0)
		RTi:SetTexture(cfg.RTc)
		self.RaidTargetIndicator = RTi
	end,

	party = function(self, ...)
		Shared(self, ...)
		Castbar(self)
		dHlight(self)
		movHlight(self)
		Portrait(self)
		SpellRange(self)
		threatHlight(self)
		HighlightAura(self)
		
		-- Unit Frame Size
		self.Health:SetHeight(cfg.UFS.Party.hh)
		self.Power:SetSize(cfg.UFS.Party.w, cfg.UFS.Party.ph)
		self.Castbar.Text:SetWidth(cfg.UFS.Party.w * 0.6)
		self.Castbar.Icon:SetSize(cfg.UFS.Party.hh, cfg.UFS.Party.hh)

		-- Status strings
		self.status:SetPoint('BOTTOMLEFT', self.Health, 'TOPLEFT', 0, 2)
		self.name:SetWidth(self:GetWidth() * 0.7)
		self.name:SetPoint('LEFT', self.status, 'RIGHT', 0, 0)
		self.name:SetJustifyH('LEFT')
		self.name:SetWordWrap(false)
		self.perhp:SetPoint('BOTTOMRIGHT', self.Health, 'TOPRIGHT', 3, 2)
		
		-- Switch element
    	self:HookScript('OnEnter', function()
			self.curhp:SetAlpha(0)
			self.movHlight:Show()
			UIFrameFadeOut(self.hp, 0.5, self.hp:GetAlpha(), 1)
			UIFrameFadeOut(self.pp, 0.5, self.pp:GetAlpha(), 1)
			UIFrameFadeOut(self.Castbar, 0, self.Castbar:GetAlpha(), 0)
			UIFrameFadeOut(self.element, cfg.delay, self.element:GetAlpha(), 0)
            UIFrameFadeOut(self.Portrait, cfg.delay, self.Portrait:GetAlpha(), 0)
    	end)
        
    	self:HookScript('OnLeave', function()
			self.curhp:SetAlpha(1)
			self.movHlight:Hide()
			UIFrameFadeIn(self.hp, 0.5, self.hp:GetAlpha(), 0)
			UIFrameFadeIn(self.pp, 0.5, self.pp:GetAlpha(), 0)
			UIFrameFadeIn(self.Castbar, cfg.delay, self.Castbar:GetAlpha(), 1)
			UIFrameFadeIn(self.element, cfg.delay, self.element:GetAlpha(), 1)
            UIFrameFadeIn(self.Portrait, cfg.delay, self.Portrait:GetAlpha(), cfg.pAlpha)
		end)
		
    	-- Auras
    	CreateBuffs(self)
    	self.Buffs:SetPoint('BOTTOMRIGHT', self.Health, 'LEFT', -36, 2)
    	self.Buffs.initialAnchor = 'BOTTOMRIGHT'
    	self.Buffs.num = 8
    	self.Buffs.size = cfg.auraSizeP
    	self.Buffs['growth-x'] = 'LEFT'
    	self.Buffs.spacing = 4
		self.Buffs:SetSize(self.Buffs.num * (self.Buffs.size + self.Buffs.spacing), self.Buffs.size + self.Buffs.spacing)
        
    	CreateDebuffs(self)
    	self.Debuffs:SetPoint('TOPRIGHT', self.Buffs, 'BOTTOMRIGHT', 0, -2)
    	self.Debuffs.initialAnchor = 'BOTTOMRIGHT'
    	self.Debuffs.num = 8
    	self.Debuffs.size = cfg.auraSizeP
    	self.Debuffs['growth-x'] = 'LEFT'
    	self.Debuffs.spacing = 4
		self.Debuffs:SetSize(self.Debuffs.num * (self.Debuffs.size + self.Debuffs.spacing), self.Debuffs.size + self.Debuffs.spacing)
		
		-- HealComm
		HealPrediction(self)

		-- Element
    	local element = CreateFrame('Frame', nil, self)
    	element:SetAllPoints(self.Health)
    	element:SetFrameLevel(self.Health:GetFrameLevel())
    	self.element = element
		
		CIi = element:CreateTexture(nil, 'OVERLAY')
		CIi:SetSize(24, 22)
		CIi:SetPoint('LEFT', self.Health, 2, 1)
		CIi:SetTexture(cfg.Cit)
		self.CombatIndicator = CIi

		LIi = element:CreateTexture(nil, 'OVERLAY')
		LIi:SetSize(20, 20)
		LIi:SetPoint('LEFT', CIi, 'RIGHT', -1, -1)
		LIi:SetTexture(cfg.Lit)
		self.LeaderIndicator = LIi
		
		MLi = element:CreateTexture(nil, 'OVERLAY')
		MLi:SetSize(18, 18)
		MLi:SetPoint('LEFT', LIi, 'RIGHT', 0, 1)
		MLi:SetTexture(cfg.MLi)
		self.MasterLooterIndicator = MLi
		
		Afk = element:CreateTexture(nil, 'OVERLAY')
		Afk:SetSize(16, 16)
		Afk:SetPoint('TOPRIGHT', self.Health, 2, 1)
		Afk:SetTexture(cfg.Afk)
		self.AfkIndicator = Afk

		Dnd = element:CreateTexture(nil, 'OVERLAY')
		Dnd:SetSize(16, 16)
		Dnd:SetPoint('TOPRIGHT', self.Health, 2, 1)
		Dnd:SetTexture(cfg.Dnd)
		self.DndIndicator = Dnd

		FDi = element:CreateTexture(nil, 'OVERLAY')
    	FDi:SetSize(18, 18)
    	FDi:SetPoint('CENTER', self.Health)
		self.FeignIndicator = FDi

		Off = element:CreateTexture(nil, 'OVERLAY')
    	Off:SetSize(24, 24)
    	Off:SetPoint('CENTER', self.Health)
    	self.OfflineIndicator = Off

    	RCi = element:CreateTexture(nil, 'OVERLAY')
    	RCi:SetSize(18, 18)
    	RCi:SetPoint('LEFT', self.Health, 'RIGHT', 4, 0)
		self.ReadyCheckIndicator = RCi

		Res = element:CreateTexture(nil, 'OVERLAY')
    	Res:SetSize(20, 20)
    	Res:SetPoint('CENTER', self.Health)
		self.ResurrectIndicator = Res

		RTi = self:CreateTexture(nil, 'OVERLAY')
		RTi:SetSize(20, 20)
		RTi:SetPoint('LEFT', self.Health, 'RIGHT', 4, 0)
		RTi:SetTexture(cfg.RTc)
		self.RaidTargetIndicator = RTi
	end,

	raid = function(self, ...)
		Shared(self, ...)
		rdHlight(self)
		movHlight(self)
		targetGlow(self)
		threatGlow(self)
		SpellRange(self)
		
		-- Unit Frame Size
		self.Health:SetHeight(cfg.UFS.Raid.hh)
    	self.Power:ClearAllPoints()
    	self.Power:SetSize(cfg.UFS.Raid.w * 0.6, cfg.UFS.Raid.ph)
    	self.Power:SetPoint('BOTTOMLEFT', self.Health, 2, 1)
		
		-- Status strings
		self.name:SetFont(cfg.NameFont, cfg.SmallFS, cfg.FontF)
    	self.name:SetPoint('RIGHT', self.Health, 2, 0)
		self.name:SetJustifyH('RIGHT')
		self:Tag(self.name, '[raidcolor][raidmissinghpname]')

		-- AuraWatch
		if cfg.AuraWatch then
			local AuraWatch = CreateFrame('Frame', nil, self)
			AuraWatch:SetAllPoints(self.Health)
			AuraWatch.hideCount = true
			AuraWatch.strictMatching = true
			AuraWatch.icons = {}
			
			local List = {}
			if filter.AuraWatch['ALL'] then
        		for _, value in pairs(filter.AuraWatch['ALL']) do
            		tinsert(List, value)
        		end
			end
			if filter.AuraWatch[playerClass] then
        		for _, value in pairs(filter.AuraWatch[playerClass]) do
            		tinsert(List, value)
        		end
			end
			if (List) then
				for _, value in pairs(List) do
					local icon = CreateFrame('Frame', nil, AuraWatch)
					icon.spellID = value[1]
					icon.anyUnit = value[5]
					if value[6] then
						icon.presentAlpha = value[6]
					else
						icon.presentAlpha = 0.6
					end
					icon.missingAlpha = 0
					icon:SetPoint(value[2], AuraWatch, value[3], value[4])
					icon:SetSize(self:GetHeight() * 0.3, self:GetHeight() * 0.3)
					AuraWatch.icons[value[1]] = icon
				end
			end
			self:HookScript('OnEnter', function()
				UIFrameFadeOut(AuraWatch, 0.5, AuraWatch:GetAlpha(), 0)
			end)
			self:HookScript('OnLeave', function()
				UIFrameFadeIn(AuraWatch, cfg.delay, AuraWatch:GetAlpha(), 1)
			end)
			
			self.AuraWatch = AuraWatch
		end

		-- Raid Debuffs
		if cfg.RaidDebuffs then
			local RaidDebuffs = CreateFrame('Frame', nil, self, 'BackdropTemplate')
			RaidDebuffs:SetSize(self.Health:GetHeight() * 0.8, self.Health:GetHeight() * 0.8)
			RaidDebuffs:SetPoint('LEFT', self.Health, 1, 1)
			RaidDebuffs:SetBackdrop(cfg.frameBD)
			RaidDebuffs:SetAlpha(0.8)
			RaidDebuffs:SetFrameLevel(self.Health:GetFrameLevel() + 1)
			RaidDebuffs.showDispelableDebuff = true
		
			RaidDebuffs.icon = RaidDebuffs:CreateTexture(nil, 'ARTWORK')
			RaidDebuffs.icon:SetPoint('TOPLEFT', RaidDebuffs, 3, -3)
			RaidDebuffs.icon:SetPoint('BOTTOMRIGHT', RaidDebuffs, -3, 3)
			
			RaidDebuffs.time = RaidDebuffs:CreateFontString(nil, 'OVERLAY')
			RaidDebuffs.time:SetFont(cfg.AuraFont, cfg.LargeFS, cfg.FontF)
			RaidDebuffs.time:SetPoint('TOPLEFT', RaidDebuffs.icon, 2, 4)
			RaidDebuffs.time:SetJustifyH('LEFT')

			RaidDebuffs.count = RaidDebuffs:CreateFontString(nil, 'OVERLAY')
			RaidDebuffs.count:SetFont(cfg.AuraFont, cfg.LargeFS, cfg.FontF)
			RaidDebuffs.count:SetPoint('BOTTOMRIGHT', RaidDebuffs.icon, 2, 0)
			RaidDebuffs.count:SetTextColor(unpack(cfg.countcolor))
			RaidDebuffs.count:SetJustifyH('RIGHT')

			self.RaidDebuffs = RaidDebuffs
		end
		
		-- HealComm
		HealPrediction(self)

		-- DebuffHighlightFilter
		self.DebuffHighlightFilter = true
		
		-- mouseover highlight
		self.movHlight:SetAllPoints(self.Health)
		self:HookScript('OnEnter', function() self.movHlight:Show() end)
		self:HookScript('OnLeave', function() self.movHlight:Hide() end)
		
		-- Element
    	local element = CreateFrame('Frame', nil, self)
    	element:SetAllPoints(self.Health)
		element:SetFrameLevel(self.Health:GetFrameLevel() + 3)
		
    	LIi = element:CreateTexture(nil, 'OVERLAY')
    	LIi:SetSize(12, 12)
    	LIi:SetPoint('TOPLEFT', self.Health, -2, 8)
    	self.LeaderIndicator = LIi
        
    	Asi = element:CreateTexture(nil, 'OVERLAY')
    	Asi:SetSize(12, 12)
    	Asi:SetPoint('TOPLEFT', self.Health, -2, 8)
		self.AssistantIndicator = Asi
		
		FDi = element:CreateTexture(nil, 'OVERLAY')
    	FDi:SetSize(16, 16)
    	FDi:SetPoint('CENTER', self.Health)
		self.FeignIndicator = FDi

		MLi = element:CreateTexture(nil, 'OVERLAY')
		MLi:SetPoint('TOPLEFT', self.Health, 8, 8)
		MLi:SetSize(13, 12)
		MLi:SetTexture(cfg.MLi)
		self.MasterLooterIndicator = MLi

		Off = element:CreateTexture(nil, 'OVERLAY')
    	Off:SetSize(22, 22)
    	Off:SetPoint('LEFT', self.Health, 2, 0)
    	self.OfflineIndicator = Off

		RRI = element:CreateTexture(nil, 'OVERLAY')
    	RRI:SetSize(11, 11)
    	RRI:SetPoint('TOPRIGHT', self.Health, 4, 7)
		self.RaidRoleIndicator = RRI
		
    	RCi = element:CreateTexture(nil, 'OVERLAY')
    	RCi:SetSize(16, 16)
    	RCi:SetPoint('CENTER', self.Health)
		self.ReadyCheckIndicator = RCi

		Res = element:CreateTexture(nil, 'OVERLAY')
    	Res:SetSize(18, 18)
    	Res:SetPoint('CENTER', self.Health)
		self.ResurrectIndicator = Res
		
		RTi = element:CreateTexture(nil, 'OVERLAY')
		RTi:SetSize(14, 14)
		RTi:SetPoint('TOP', self.Health, -10, 8)
		RTi:SetTexture(cfg.RTc)
		self.RaidTargetIndicator = RTi
	end
}

---------------------------------------
-- Register style(s) and spawn units --
---------------------------------------
for unit,layout in next, UnitSpecific do
	-- Capitalize the unit name, so it looks better.
	oUF:RegisterStyle('Nihlathak - ' .. unit:gsub('^%l', string.upper), layout)
end

-- A small helper to change the style into a unit specific, if it exists.
local Spawn = function(self, unit, ...)
	if(UnitSpecific[unit]) then
		self:SetActiveStyle('Nihlathak - ' .. unit:gsub('^%l', string.upper))
	else
		self:SetActiveStyle('Nihlathak')
	end

	local object = self:Spawn(unit)
	object:SetPoint(...)
	return object
end

oUF:Factory(function(self)
	local player = Spawn(self, 'player', unpack(cfg.UFP.Player))
	local pet = Spawn(self, 'pet', unpack(cfg.UFP.Pet))
	local target = Spawn(self, 'target', unpack(cfg.UFP.Target))
	local targettarget = Spawn(self, 'targettarget', unpack(cfg.UFP.TT))
	
	if cfg.MTMAFrame then
		self:SetActiveStyle('Nihlathak - Tank')
		local tank = self:SpawnHeader(nil, nil, 'raid', 
			'showRaid', true,
			'groupFilter', 'MAINTANK, MAINASSIST',
			'groupingOrder', 'MAINTANK, MAINASSIST',
			'groupBy', 'ROLE',
			'yOffset', -35,
			'oUF-initialConfigFunction', ([[
				self:SetWidth(%d)
				self:SetHeight(%d)
			]]):format(cfg.UFS.Tank.w, cfg.UFS.Tank.h)
		)
		tank:SetPoint(unpack(cfg.UFP.Tank))
	end

	if cfg.PartyFrame then
		self:SetActiveStyle('Nihlathak - Party')
		local party = self:SpawnHeader(nil, nil, 'party',
			'showParty', true,
			'yOffset', -40,
			'groupingOrder', 'TANK, HEALER, DAMAGER, NONE',
			'groupBy', 'ROLE',
			'oUF-initialConfigFunction', ([[
				self:SetWidth(%d)
				self:SetHeight(%d)
			]]):format(cfg.UFS.Party.w, cfg.UFS.Party.h)
		)
		party:SetPoint(unpack(cfg.UFP.Party))
	end

	if cfg.RaidFrame then
		self:SetActiveStyle('Nihlathak - Raid')
		local raid = {}
		for group = 1, _G.NUM_RAID_GROUPS do
			raid[group] = self:SpawnHeader(nil, nil, 'raid',
				'showRaid', true,
				'groupFilter', group,
				'maxColumns', 5,
				'unitsPerColumn', 1,
				'columnSpacing', 8,
				'columnAnchorPoint', 'LEFT',
				'oUF-initialConfigFunction', ([[
					self:SetWidth(%d)
					self:SetHeight(%d)
				]]):format(cfg.UFS.Raid.w, cfg.UFS.Raid.h)
			)
			if (group == 1) then
				raid[group]:SetPoint(unpack(cfg.UFP.Raid))
			else
				raid[group]:SetPoint('TOPLEFT', raid[group - 1], 'BOTTOMLEFT', 0, -8)
			end
		end
	end
end)

-- target select sound
local handler = CreateFrame('Frame')
handler:SetScript('OnEvent', function(self, event, ...)
	self[event](self, ...)
end)

function handler:PLAYER_TARGET_CHANGED()
	if UnitExists('target') and not IsReplacingUnit() then
		if UnitIsEnemy('target', 'player') then
			PlaySound(SOUNDKIT.IG_CREATURE_AGGRO_SELECT)
		elseif UnitIsFriend('target', 'player') then
			PlaySound(SOUNDKIT.IG_CHARACTER_NPC_SELECT)
		else
			PlaySound(SOUNDKIT.IG_CREATURE_NEUTRAL_SELECT)
		end
	else
		PlaySound(SOUNDKIT.INTERFACE_SOUND_LOST_TARGET_UNIT)
	end
end
handler:RegisterEvent('PLAYER_TARGET_CHANGED')

-------------
-- The End --
-------------